#include "lossless_recorder.h"

Lossless_Recorder::Lossless_Recorder(QObject *parent) :
        QObject(parent)
      ,mutex(NULL)
      ,sws_ctx(NULL)
      ,currently_recording(false)
      ,input_avpicture(NULL)
      ,output_avframe(NULL)
      ,audio_frame(NULL)
      ,input_pixel_format(AV_PIX_FMT_BGRA)
      ,output_pixel_format(AV_PIX_FMT_YUV444P)

{
        av_register_all();
        avformat_network_init();
}

Lossless_Recorder::~Lossless_Recorder() {
        clean_up_ffmpeg_crap();
}

/*
 * Copyright (c) 2003 Fabrice Bellard
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

/**
 * @file
 * libavformat API example.
 *
 * Output a media file in any supported libavformat format.
 * The default codecs are used.
 * @example doc/examples/muxing.c
 */

/* Add an output stream. */
AVStream *Lossless_Recorder::add_stream(AVFormatContext *oc, AVCodec **codec, enum AVCodecID codec_id) {
        AVCodecContext *c;
        AVStream *st;

        /* find the encoder */
        *codec = avcodec_find_encoder(codec_id);
        if (!(*codec)) {
                qDebug() << this << tr("Could not find encoder for '%1'").arg(avcodec_get_name(codec_id));
                emit error_with_string(tr("Could not find encoder for '%1'").arg(avcodec_get_name(codec_id)));
                return NULL;
        }

        st = avformat_new_stream(oc, *codec);
        if (!st) {
                qDebug() << this << tr("Could not allocate stream");
                emit error_with_string(tr("Could not allocate stream"));
                return NULL;
        }
        st->id = oc->nb_streams-1;
        c = st->codec;

        switch ((*codec)->type) {
        case AVMEDIA_TYPE_AUDIO:
                st->id = 1;
                c->sample_fmt  = AV_SAMPLE_FMT_S16;
                c->sample_rate = audio_sample_rate_bits;
                c->channels    = audio_number_of_channels;
                //c->channel_layout = c->channels == 2 ? AV_CH_LAYOUT_STEREO : AV_CH_LAYOUT_MONO;
                break;

        case AVMEDIA_TYPE_VIDEO:
                avcodec_get_context_defaults3(c, *codec);
                c->codec_id = codec_id;

                c->width    = video_width;
                c->height   = video_height;
                c->time_base.num = 1000;
                c->time_base.den = (int)(video_framerate * 1000.0);
                video_frame_time_ms = ((double)c->time_base.num / (double)c->time_base.den) * 1000.0;
                c->gop_size      = 60; /* emit one intra frame every 60 frames at most */
                c->pix_fmt       = output_pixel_format;
                c->thread_count = 0;
                break;

        default:
                break;
        }

        /* Some formats want stream headers to be separate. */
        if (oc->oformat->flags & AVFMT_GLOBALHEADER)
                c->flags |= CODEC_FLAG_GLOBAL_HEADER;

        return st;
}

void Lossless_Recorder::open_audio(AVFormatContext *oc, AVCodec *codec, AVStream *st) {
        Q_UNUSED(oc);
        AVCodecContext *c;
        int ret;

        c = st->codec;

        /* open it */
        if (mutex) mutex->lock();
        ret = avcodec_open2(c, codec, NULL);
        if (mutex) mutex->unlock();
        if (ret < 0) {
                char ffmpeg_error_p[AV_ERROR_MAX_STRING_SIZE];
                av_make_error_string(ffmpeg_error_p, AV_ERROR_MAX_STRING_SIZE, ret);
                QString error_string = tr("Could not open audio codec: %1").arg(ffmpeg_error_p);
                emit error_with_string(error_string);
                return;
        }

        if (c->codec->capabilities & CODEC_CAP_VARIABLE_FRAME_SIZE)
                audio_samples_per_channel = 10000;
        else
                audio_samples_per_channel = c->frame_size;

        audio_frame_size = audio_samples_per_channel * av_get_bytes_per_sample(c->sample_fmt) * c->channels;

        samples = (int16_t *)av_malloc(audio_frame_size);
        if (!samples) {
                qDebug() << this << tr("Could not allocate audio samples buffer");
                emit error_with_string(tr("Could not allocate audio samples buffer"));
                return;
        }

        audio_frame = avcodec_alloc_frame();
        audio_frame->nb_samples = audio_samples_per_channel;
        avcodec_fill_audio_frame(audio_frame, c->channels, c->sample_fmt,
                                 (uint8_t *)samples,
                                 audio_frame_size, 1);
}

void Lossless_Recorder::write_audio_frame(AVFormatContext *oc, AVStream *st) {
        //qDebug() << "write_audio_frame()";
        AVCodecContext *c = st->codec;

        if (audio_data.size() < audio_frame_size) {
                qDebug() << this << "audio underflow!";
                return;
        }

        uint16_t *in = (uint16_t *)audio_data.data();

        if (0) { //hack to transform to planar output
                uint16_t *left_channel = (uint16_t *)audio_frame->extended_data[0];
                uint16_t *right_channel = (uint16_t *)audio_frame->extended_data[1];
                int top = audio_frame_size/2/2;
                for (int i = 0; i < top; ++i, in += 2) {
                        left_channel[i] = in[0];
                        right_channel[i] = in[1];
                }
        } else {
                memcpy(audio_frame->data[0], in, audio_frame_size);
        }

        audio_data = audio_data.mid(audio_frame_size);

        int got_packet, ret;

        AVPacket pkt;
        av_init_packet(&pkt);
        pkt.size = 0;
        pkt.data = 0;


        ret = avcodec_encode_audio2(c, &pkt, audio_frame, &got_packet);
        if (ret < 0) {
                char ffmpeg_error_p[AV_ERROR_MAX_STRING_SIZE];
                av_make_error_string(ffmpeg_error_p, AV_ERROR_MAX_STRING_SIZE, ret);
                QString error_string = tr("Error encoding audio frame: %1").arg(ffmpeg_error_p);
                emit error_with_string(error_string);
                return;
        }

        if (!got_packet) {
                return;
        }

        pkt.stream_index = st->index;

        /* Write the compressed frame to the media file. */
        ret = av_interleaved_write_frame(oc, &pkt);
        if (ret != 0) {
                char ffmpeg_error_p[AV_ERROR_MAX_STRING_SIZE];
                av_make_error_string(ffmpeg_error_p, AV_ERROR_MAX_STRING_SIZE, ret);
                QString error_string = tr("Error while writing audio frame: %1").arg(ffmpeg_error_p);
                emit error_with_string(error_string);
                return;
        }
        ++audio_frames_encoded;
}

void Lossless_Recorder::close_audio(AVFormatContext *oc, AVStream *st) {
        Q_UNUSED(oc);
        if (mutex) mutex->lock();
        avcodec_close(st->codec);
        if (mutex) mutex->unlock();

        av_free(samples);
}

void Lossless_Recorder::open_video(AVFormatContext *oc, AVCodec *codec, AVStream *st) {
        Q_UNUSED(oc);
        int ret;
        AVCodecContext *c = st->codec;

        AVDictionary *opts = NULL;
        QByteArray video_preset_ba = video_preset.toLocal8Bit();
        const char *video_preset_pointer = video_preset_ba.constData();
        av_dict_set(&opts, "preset", video_preset_pointer, 0); //this automatically sets the profile (20130313)
        av_dict_set(&opts, "crf", "0", 0); //lossless

        /* open the codec */
        if (mutex) mutex->lock();
        ret = avcodec_open2(c, codec, &opts);
        if (mutex) mutex->unlock();
        if (ret < 0) {
                char ffmpeg_error_p[AV_ERROR_MAX_STRING_SIZE];
                av_make_error_string(ffmpeg_error_p, AV_ERROR_MAX_STRING_SIZE, ret);
                QString error_string = tr("Could not open video codec: %1").arg(ffmpeg_error_p);
                emit error_with_string(error_string);
                return;
        }
}

void Lossless_Recorder::write_video_frame(AVFormatContext *oc, AVStream *st) {
        //qDebug() << "write_video_frame()";
        if (!video_frames.size()) {
                qDebug() << this << "video underflow!";
                return;
        }

        int ret;
        AVCodecContext *c = st->codec;

        QImage input_frame = video_frames.takeFirst();
        //input_frame = input_frame.convertToFormat(QImage::Format_RGB888);

        memcpy(input_avpicture->data[0], input_frame.bits(), input_frame.byteCount());
        sws_scale(sws_ctx,
                  input_avpicture->data, input_avpicture->linesize, 0, input_frame.height(),
                  output_avframe->data, output_avframe->linesize);

        /* encode the image */
        AVPacket pkt;
        int got_output;

        av_init_packet(&pkt);
        pkt.data = NULL;    // packet data will be allocated by the encoder
        pkt.size = 0;

        ret = avcodec_encode_video2(c, &pkt, output_avframe, &got_output);
        ++output_avframe->pts;
        if (ret < 0) {
                char ffmpeg_error_p[AV_ERROR_MAX_STRING_SIZE];
                av_make_error_string(ffmpeg_error_p, AV_ERROR_MAX_STRING_SIZE, ret);
                QString error_string = tr("Error encoding video frame: %1").arg(ffmpeg_error_p);
                emit error_with_string(error_string);
                return;
        }

        /* If size is zero, it means the image was buffered. */
        if (got_output) {
                if (c->coded_frame->key_frame)
                        pkt.flags |= AV_PKT_FLAG_KEY;

                pkt.stream_index = st->index;

                /* Write the compressed frame to the media file. */
                ret = av_interleaved_write_frame(oc, &pkt);
        } else {
                ret = 0;
        }
        if (ret != 0) {
                char ffmpeg_error_p[AV_ERROR_MAX_STRING_SIZE];
                av_make_error_string(ffmpeg_error_p, AV_ERROR_MAX_STRING_SIZE, ret);
                QString error_string = tr("Error while writing video frame: %1").arg(ffmpeg_error_p);
                emit error_with_string(error_string);
                return;
        }
        ++video_frames_encoded;
}

void Lossless_Recorder::close_video(AVFormatContext *oc, AVStream *st) {
        Q_UNUSED(oc);
        if (mutex) mutex->lock();
        avcodec_close(st->codec);
        if (mutex) mutex->unlock();
}

void Lossless_Recorder::start(QString url, int width, int height, double framerate, QString preset, int new_audio_sample_rate_bits, int new_audio_number_of_channels) {
        if (currently_recording) return;
        video_width = width;
        video_height = height;
        video_framerate = framerate;
        video_preset = preset;
        audio_sample_rate_bits = new_audio_sample_rate_bits;
        audio_number_of_channels = new_audio_number_of_channels;

        video_frames.clear();
        audio_data.clear();

        sws_flags = SWS_LANCZOS;

        filename_ba = url.toLocal8Bit();
        filename = filename_ba.constData();

        avformat_alloc_output_context2(&oc, NULL, "avi", filename); //why is "avi" required here? (20130513)
        Q_ASSERT(oc);
        fmt = oc->oformat;

        video_st = NULL;
        audio_st = NULL;

        if (fmt->video_codec != AV_CODEC_ID_NONE) {
                fmt->video_codec = AV_CODEC_ID_H264;
                video_st = add_stream(oc, &video_codec, fmt->video_codec);
        }
        if (fmt->audio_codec != AV_CODEC_ID_NONE) {
                fmt->audio_codec = AV_CODEC_ID_PCM_S16LE;
                audio_st = add_stream(oc, &audio_codec, fmt->audio_codec);
        }

        /* Now that all the parameters are set, we can open the audio and
     * video codecs and allocate the necessary encode buffers. */
        if (video_st)
                open_video(oc, video_codec, video_st);
        if (audio_st)
                open_audio(oc, audio_codec, audio_st);

        av_dump_format(oc, 0, filename, 1);

        /* open the output file, if needed */
        if (!(fmt->flags & AVFMT_NOFILE)) {
                int ret = avio_open(&oc->pb, filename, AVIO_FLAG_WRITE);
                if (ret < 0) {
                        char ffmpeg_error_p[AV_ERROR_MAX_STRING_SIZE];
                        av_make_error_string(ffmpeg_error_p, AV_ERROR_MAX_STRING_SIZE, ret);
                        QString error_string = tr("Could not open: %1").arg(ffmpeg_error_p);
                        emit error_with_string(error_string);
                        return;
                }
        }

        /* Write the stream header, if any. */
        int ret = avformat_write_header(oc, NULL);
        if (ret < 0) {
                char ffmpeg_error_p[AV_ERROR_MAX_STRING_SIZE];
                av_make_error_string(ffmpeg_error_p, AV_ERROR_MAX_STRING_SIZE, ret);
                QString error_string = tr("Error occurred when opening output file: %1").arg(ffmpeg_error_p);
                emit error_with_string(error_string);
                return;
        }

        sws_ctx = sws_getContext(video_width, video_height, input_pixel_format,
                                 video_width, video_height, output_pixel_format,
                                 sws_flags, NULL, NULL, NULL);
        if (!sws_ctx) {
                emit error_with_string(tr("Could not initialize the conversion context"));
                return;
        }

        input_avpicture = (AVPicture *)av_malloc(sizeof(AVPicture));
        avpicture_alloc(input_avpicture, input_pixel_format, video_width, video_height);
        output_avframe = avcodec_alloc_frame();
        avpicture_alloc(&dst_picture, output_pixel_format, video_width, video_height);
        /* copy data and linesize picture pointers to frame */
        *((AVPicture *)output_avframe) = dst_picture;
        output_avframe->pts = 0;
        audio_frames_encoded = video_frames_encoded = 0;
        session_timer.start();

        currently_recording = true;
}

void Lossless_Recorder::add_video_frame(QImage frame) {
        if (!currently_recording) return;
        Q_ASSERT(frame.width() == video_width);
        Q_ASSERT(frame.height() == video_height);

        video_frames << frame;

        return record_frames();
}

void Lossless_Recorder::add_audio_data(QByteArray data) {
        if (!currently_recording) return;

        audio_data.append(data);

        return record_frames();
}

void Lossless_Recorder::record_frames() {
        if (!currently_recording) return;
        //qDebug() << "record_frames(); audio_data.size()" << audio_data.size() << "video_frames.size()" << video_frames.size();
        while (audio_data.size() >= audio_frame_size && video_frames.size()) {
                double video_pts = (double)video_st->pts.val * video_st->time_base.num / video_st->time_base.den;
                double audio_pts = (double)audio_st->pts.val * audio_st->time_base.num / audio_st->time_base.den;
                if (audio_pts < video_pts)
                        write_audio_frame(oc, audio_st);
                else
                        write_video_frame(oc, video_st);
        }


        //
        return;
        //


        //do not allow the audio to run more than 3 frames behind - we are more slack here than with the video because we prefer to drop audio frames (20111009)
        if (audio_data.size() > audio_frame_size*3) {
                audio_data = audio_data.right(audio_frame_size*2);
                qDebug() << this << "dropped audio frames";
        }

        //do not allow the video to run more than 1 frame behind the audio (20111013)
        if (video_frames.size() > 1 && audio_data.size() < audio_frame_size) {
                int audio_frames_per_video_frame = (audio_sample_rate_bits * 2 * 2) / video_frame_time_ms / audio_frame_size;
                for (int i = 0; i < video_frames.size() - 1; i++) { //this may come up to one frame short (20111013)
                        qDebug() << this << "adding" << audio_frames_per_video_frame << "audio frames to audio_data";
                        for (int j = 0; j < audio_frames_per_video_frame; j++) {
                                QByteArray silence(audio_frame_size, 0);
                                audio_data.append(silence);
                        }
                }
        }
}

void Lossless_Recorder::clean_up_ffmpeg_crap() {
        if (sws_ctx) {
                sws_freeContext(sws_ctx);
                sws_ctx = NULL;
        }
        if (input_avpicture) {
                avpicture_free(input_avpicture);
                input_avpicture = NULL;
        }
        if (output_avframe) {
                avpicture_free(&dst_picture);
                avcodec_free_frame(&output_avframe);
                output_avframe = NULL;
        }
        if (audio_frame) {
                avcodec_free_frame(&audio_frame);
                audio_frame = NULL;
        }
}

void Lossless_Recorder::stop() {
        if (!currently_recording) return;

        record_frames();

        clean_up_ffmpeg_crap();

        /* Write the trailer, if any. The trailer must be written before you
     * close the CodecContexts open when you wrote the header; otherwise
     * av_write_trailer() may try to use memory that was freed on
     * av_codec_close(). */
        av_write_trailer(oc);

        /* Close each codec. */
        if (video_st)
                close_video(oc, video_st);
        if (audio_st)
                close_audio(oc, audio_st);

        /* Free the streams. */
        for (uint i = 0; i < oc->nb_streams; i++) {
                av_freep(&oc->streams[i]->codec);
                av_freep(&oc->streams[i]);
        }

        if (!(fmt->flags & AVFMT_NOFILE))
                /* Close the output file. */
                avio_close(oc->pb);

        /* free the stream */
        av_free(oc);

        currently_recording = false;
}
